"""Decorator to log metrics to Azure Application Insights.

An example of how to use this decorator:

```python
from functown import InsightsMetrics

@InsightsMetrics(
    instrumentation_key="...",
    metrics=[
        MetricSpec(
            name="my_metric",
            description="My metric",
            unit="count",
            aggregation_type=MetricType.COUNTER,
            dtype=int,
        ),
    ],
)
def my_function(metrics):
    # ...
    metrics["my_metric"].record(1)
    # ...
```

Copyright (c) 2023, Felix Geilert
"""

import logging
from typing import List, Callable, Dict, Any

from opencensus.ext.azure import metrics_exporter

from .base import InsightsDecorator
from .metrics import MetricHandler, MetricSpec, MetricUseMode


class InsightsMetrics(InsightsDecorator):
    """Decorator to log metrics to Azure Application Insights.

    This will add the `metrics` object to the function call.

    Args:
        instrumentation_key (str): The instrumentation key for the Application Insights.
        metrics (List[MetricSpec]): A list of metrics to be logged.
        enable_perf_metrics (bool): Whether to enable performance metrics to be send
            to Application Insights. Defaults to `True`.
        enable_name_column (bool): Whether to enable the `__name` column is
            automatically added to the metrics. Defaults to `False`.
        global_columns (Dict[str, Any]): A dictionary of columns that will be added
            to all metrics. Defaults to `None`.
        hard_fail (bool): Whether to hard fail if the metrics cannot be created.
            Defaults to `True`.
        flush_seconds (float): Interval (seconds) in which the metrics are flushed
            to Application Insights. Defaults to `15`.
    """

    def __init__(
        self,
        instrumentation_key: str,
        metrics: List[MetricSpec],
        callback: Callable[[metrics_exporter.Envelope], None] = None,
        enable_perf_metrics: bool = True,
        enable_name_column: bool = False,
        global_columns: Dict[str, Any] = None,
        mode: MetricUseMode = MetricUseMode.REUSE,
        flush_seconds: float = 15,
        **kwargs,
    ):
        super().__init__(instrumentation_key, added_kw=["metrics"], **kwargs)

        self.metric_specs = metrics
        self.callback = callback
        self.perf_metrics = enable_perf_metrics
        self.enable_name_column = enable_name_column
        self.global_columns = global_columns
        self.mode = mode
        self.flush_sec = flush_seconds
        self._handler = None

    def run(self, func, *args, **kwargs):
        try:
            # generate the handler
            self._handler = MetricHandler(self.enable_name_column)
            self._handler.create_metrics(
                self.metric_specs,
                mode=self.mode,
                global_columns=self.global_columns,
            )
            kwargs["metrics"] = self._handler

            # register exporter for the metrics
            if self.instrumentation_key is not None:
                self._handler.connect_insights(
                    self.instrumentation_key,
                    self.callback,
                    enable_standard_metrics=self.perf_metrics,
                    flush_sec=self.flush_sec,
                )
        except Exception as ex:
            logging.error(f"Failed to create Metrics Logger: {ex}")
            raise ex

        return func(*args, **kwargs)

    def teardown(self, *args, **kwargs):
        if self._handler is not None:
            self._handler.shutdown()



__version__ = "1.1.1"

import logging

try:
    from . import utils
    from . import errors
    from .errors import ErrorHandler
    from .args import RequestArgHandler, ArgsHandler
except Exception as ex:
    logging.error(f"Not all dependencies installed: {ex}")

try:
    from . import auth
    from .auth import AuthHandler
except Exception as ex:
    logging.error(f"Unable to load auth: {ex}")

try:
    from . import serialization
except Exception as ex:
    logging.error(f"Unable to load serialization: {ex}")

try:
    from . import insights
    from .insights import (
        InsightsLogs,
        InsightsMetrics,
        InsightsEvents,
        InsightsTracer,
        Insights,
    )
except Exception as ex:
    logging.error(f"Unable to load metrics: {ex}")



import logging


class LogListHandler(logging.Handler):
    """List handler that retrives a reference to a list to append logs to

    Usage:

    ```python
    log_list = []
    logger = logging.getLogger()
    logger.addHandler(LogListHandler(log_list))

    # do stuff
    # log_list now contains all logs
    ```

    """

    def __init__(self, log_list):
        logging.Handler.__init__(self)
        self.log_list = log_list

    def emit(self, record):
        # record.message is the log message
        self.log_list.append(record.msg)



"""Implements a decorator that allows to stack decorators.

Copyright (c) 2023 Felix Geilert
"""

from typing import List

from .base_decorator import BaseDecorator


class StackDecorator(BaseDecorator):
    """A decorator that allows to stack decorators.

    This can also execute custom functions by overriding the run_pre_hook and
    run_post_hook methods.

    Args:
        stack (List[BaseDecorator]): A list of decorators to apply.
        added_kw (List[str]): A list of keywords that are added by the decorator.
    """

    def __init__(
        self,
        stack: List[BaseDecorator],
        added_kw: List[str] = None,
        **kwargs,
    ):
        # set and expand
        self.__decs = stack
        kws = added_kw or []
        for dec in self.__decs:
            kws.extend(dec.added_kw)

        # init
        super().__init__(None, added_kw=kws, **kwargs)

    def run_pre_hook(self, func, *args, **kwargs):
        """Child classes can override this method to execute a pre hook."""
        return func

    def run_post_hook(self, func, *args, **kwargs):
        """Child classes can override this method to execute a post hook."""
        return func

    def run(self, func, *args, **kwargs):
        # execute pre hook
        func = self.run_pre_hook(func, *args, **kwargs)

        # apply all decorators
        for dec in self.__decs:
            func = dec(func)

        # execute post hook
        func = self.run_post_hook(func, *args, **kwargs)

        return func(*args, **kwargs)



"""
Function decorator that should make error handling easier

Copyright (c) 2023, Felix Geilert
"""

from types import TracebackType
from typing import List
import os
import sys
import json
import logging

from azure.functions import HttpResponse

from functown.utils import BaseDecorator, LogListHandler
from .errors import RequestError


class ErrorHandler(BaseDecorator):
    """Decorator to handle errors inside of Azure Function Code.

    This might add two additional parameters to the inner function signature:
    - logger: logging.Logger (if enable_logger is True)
    - logs: List[str] (if return_logs is True)

    Args:
        debug (bool, optional): If True, the error will be returned to the user.
            Otherwise it will just send `general_error_msg` in a 500. Defaults to True.
        log_all_errors (bool, optional): If True, all errors will be send to the logger.
            Defaults to False.
        return_errors (bool, optional): If True, the error will be returned to the user.
            This means the response is a json object. Defaults to False.
        enable_logger (bool, optional): If True, a logger will be added to the function.
            The logger will be passed through as `logger` param. Defaults to False.
        return_logs (bool, optional): If True, the logs will be returned to the user.
            This only happens if `debug` is enabled. Defaults to False.
        clean_logger (bool, optional): If True, loggers from outer layers will be
            overwritten. Defaults to True.
        general_error_msg (str, optional): The message that will be returned to the user
            in the case of a general exception (otherwise returns RequestError.msg).
            Defaults to "This function executed unsuccessfully.".
    """

    def __init__(
        self,
        debug: bool = True,
        log_all_errors: bool = False,
        return_errors: bool = False,
        enable_logger: bool = False,
        return_logs: bool = False,
        clean_logger: bool = True,
        general_error_msg: str = "This function executed unsuccessfully.",
        **kwargs,
    ):
        super().__init__(None, added_kw=["logs", "logger"], **kwargs)

        # set error handling
        self.debug = debug
        self.log_all_errors = log_all_errors
        self.return_errors = return_errors
        self.enable_logger = enable_logger
        self.return_logs = return_logs
        self.clean_logger = clean_logger
        self.general_error_msg = general_error_msg

    def _send_response(
        self,
        ex: Exception,
        code: int,
        msg: str = None,
        logs: List[str] = None,
        log_error: bool = False,
    ):
        msg = msg if msg else f"Got error: {type(ex)}"
        mime = "text/plain"

        # get additional data
        exc_type, exc_obj, exc_tb = sys.exc_info()
        names = self._get_trace(exc_tb)

        # validate error type to return
        if self.return_errors is True:
            # update mime type
            mime = "application/json"
            msg = {"user_message": msg}

            # check if logs are set
            if logs is not None:
                msg["logs"] = logs

            # add exception and trace to message
            msg["type"] = str(exc_type)
            msg["value"] = str(exc_obj)
            msg["trace"] = names

            # encode for return
            msg = json.dumps(msg)

        # check for logs
        if log_error is True:
            exc_obj = ex.msg if hasattr(ex, "msg") else exc_obj
            logging.error(f"Error ({exc_type}): {str(ex)} (Obj: {exc_obj})")
            logging.error("Trace:")
            for name in names:
                logging.error(f"- {name}")

        return HttpResponse(msg, status_code=code, mimetype=mime)

    def _get_trace(self, tb: TracebackType) -> List[str]:
        names = []
        while tb:
            file_name = os.path.split(tb.tb_frame.f_code.co_filename)[1]
            fct_line_no = tb.tb_frame.f_code.co_firstlineno
            act_line_no = tb.tb_frame.f_lineno
            names.append(
                f"{file_name}:{fct_line_no}:{act_line_no} - Vars: {tb.tb_frame.f_code.co_varnames}"
            )
            tb = tb.tb_next
        return names

    def run(self, func, *args, **kwargs):
        """Decorator to handle errors

        Note: Since this decorator returns response from the function it is advised to use it
        as the last (outermost) decorator
        """
        logs = None

        # check if logger should be created
        if self.enable_logger:
            try:
                # check for logger
                logger = self._create_logger(self.clean_logger, *args, **kwargs)

                # add list system to logger
                if self.return_logs is True:
                    logs = []
                    logger.addHandler(LogListHandler(logs))
                    kwargs["logs"] = logs

                # add logger to kwargs
                if logger is not None:
                    kwargs["logger"] = logger
            except Exception as ex:
                logging.error(f"Failed to get logger: {ex}")
                raise ex
        else:
            logger = logging

        # check if outermost
        if self.is_first_decorator is False:
            logger.warning(
                "It is advised to use ErrorHandler as the first decorator, "
                f"it is currently at the {self.level} level. "
                f"(Should be {self.max_level})"
            )

        # execute main loop
        try:
            return func(*args, **kwargs)
        except RequestError as ex:
            logger.error(ex.__class__.__name__)
            return self._send_response(
                ex,
                ex.code,
                msg=ex.msg,
                logs=logs,
                log_error=self.log_all_errors,
            )
        except Exception as ex:
            logger.error("General Error")
            return self._send_response(
                ex,
                500,
                msg=self.general_error_msg,
                logs=logs,
                log_error=self.debug or self.log_all_errors,
            )



"""Callbacks for filtering and modifying Azure Application Insights."""

import os
import sys
from typing import Callable

from opencensus.ext.azure.log_exporter import Envelope


def modify_system_info(envelop: Envelope) -> bool:
    """Appends System information (os, python, cpu, memory) to the message.

    Args:
        envelop (Envelope): The telemetry item.

    Returns:
        bool: Whether to keep the item.
    """
    system_info = (
        f"System: {sys.platform} {sys.version} {os.cpu_count()} {os.getloadavg()}"
    )
    envelop.data.base_data.message = f"{envelop.data.base_data.message} {system_info}"

    return True


def filter_debug(envelop: Envelope) -> bool:
    """Filter to remove debug messages from Azure Application Insights.

    Args:
        envelop (Envelope): The telemetry item.

    Returns:
        bool: Whether to keep the item.
    """
    if envelop.data.base_data.severity_level == 1:
        return False
    return True


def create_filter_ids(cur_id, ids: list) -> Callable[[Envelope], bool]:
    """Create a filter to remove specific ids from Azure Application Insights.

    Args:
        cur_id (str): The current id of the function.
        ids (list): A list of ids to remove.

    Returns:
        Callable[[Envelope], bool]: A filter function.
    """

    def filter_ids(envelop: Envelope) -> bool:
        """Filter to remove specific ids from Azure Application Insights.

        Args:
            envelop (Envelope): The telemetry item.

        Returns:
            bool: Whether to keep the item.
        """
        if cur_id in ids:
            return False
        return True

    return filter_ids


def create_filter_keywords(keywords: list) -> Callable[[Envelope], bool]:
    """Create a filter to remove specific keywords from Azure Application Insights.

    Args:
        keywords (list): A list of keywords to remove.

    Returns:
        Callable[[Envelope], bool]: A filter function.
    """

    def filter_keywords(envelop: Envelope) -> bool:
        """Filter to remove specific keywords from Azure Application Insights.

        Args:
            envelop (Envelope): The telemetry item.

        Returns:
            bool: Whether to keep the item.
        """
        for keyword in keywords:
            if keyword in envelop.data.base_data.message:
                return False
        return True

    return filter_keywords



"""Decorator to send events to Application Insights.

This also provides a bunch of examples of callback_functions that can modify filter values.
It is also possible to run a custom compose of these filters and modifiers:

```python
def my_filter(envelop: Envelope) -> bool:
    if debug_filter(envelop):
        return False
    modify_system_info(envelop)
    return True
```

Examples how to use events:

```python
# basic event
events.info("This is an event message")
# event with extra data
events.info("This is an event message", extra={"key": "value"})
```

For a full guide on how to use the event logger, see:
https://learn.microsoft.com/en-us/azure/azure-monitor/app/opencensus-python

Copyright (c) 2023, Felix Geilert
"""

import logging
from typing import Callable

from opencensus.ext.azure.log_exporter import (
    AzureEventHandler,
    Envelope,
)

from .base import InsightsDecorator


class InsightsEvents(InsightsDecorator):
    """Decorator to send events to Application Insights.

    This will add the `events` object to the function call.

    Args:
        instrumentation_key (str): The instrumentation key for the Application Insights.
        callback (Callable[[Envelope], bool]): A callback function that is called for
            every telemetry item. It can act as a filter (by returning False) or modify
            the telemetry item (by returning True end modifying the reference to envelop
            passed).
        clean_logger (bool): Defines if existing `logger` object is overwritten.
            Defaults to False.
    """

    def __init__(
        self,
        instrumentation_key: str,
        callback: Callable[[Envelope], bool] = None,
        clean_logger: bool = False,
        **kwargs,
    ):
        super().__init__(instrumentation_key, added_kw=["events"], **kwargs)

        self.callback = callback
        self.clean_logger = clean_logger

    def run(self, func, *args, **kwargs):
        try:
            events = self._create_logger(self.clean_logger, "events", *args, **kwargs)

            if self.instrumentation_key is not None:
                event_handler = AzureEventHandler(
                    connection_string=f"InstrumentationKey={self.instrumentation_key}"
                )
                if self.callback is not None:
                    event_handler.add_telemetry_processor(self.callback)
                events.addHandler(event_handler)
            kwargs["events"] = events
        except Exception as ex:
            logging.error(f"Failed to create Insights Events Logger: {ex}")
            raise ex

        return func(*args, **kwargs)



"""Builds a logging handler for application insights.

This also provides a bunch of examples of callback_functions that can modify filter values.
It is also possible to run a custom compose of these filters and modifiers:

```python
def my_filter(envelop: Envelope) -> bool:
    if debug_filter(envelop):
        return False
    modify_system_info(envelop)
    return True
```

Examples how to use logger:

```python
# basic trace
logger.info("This is a log message")
# trace with extra data
logger.info("This is a log message", extra={"key": "value"})
# trace exception
ex = ...
logger.exception("Log found exception", exc_info=ex)
```

For a full guide on how to use the logger, see:
https://learn.microsoft.com/en-us/azure/azure-monitor/app/opencensus-python

Copyright (c) 2023, Felix Geilert
"""

import logging
from typing import Callable

from opencensus.ext.azure.log_exporter import (
    AzureLogHandler,
    Envelope,
)

from .base import InsightsDecorator


class InsightsLogs(InsightsDecorator):
    """Decorator that provides a sending interface to the Application Insights.

    This will add the `logger` object to the function call.

    Note that some default `callback` functions are defined in
    `functown.insights.callbacks`.

    Args:
        instrumentation_key (str): The instrumentation key for the Application Insights.
        enable_logger (bool): Whether to enable the logger. Defaults to True.
        send_basics (bool): Whether to send basic metrics about the function.
            Defaults to False.
        callback (Callable[[Envelope], bool]): A callback function that is called for
            every telemetry item. It can act as a filter (by returning False) or modify
            the telemetry item (by returning True end modifying the reference to envelop
            passed).
        clean_logger (bool): Defines if existing `logger` object is overwritten.
            Defaults to False.
    """

    def __init__(
        self,
        instrumentation_key: str,
        send_basics: bool = True,
        callback: Callable[[Envelope], bool] = None,
        clean_logger: bool = False,
        **kwargs,
    ):
        super().__init__(instrumentation_key, added_kw=["logger"], **kwargs)

        self.send_basics = send_basics
        self.callback = callback
        self.clean_logger = clean_logger

    def run(self, func, *args, **kwargs):
        try:
            # check for logger
            logger = self._create_logger(self.clean_logger, "logger", *args, **kwargs)

            # create azure handler
            if self.instrumentation_key is not None:
                log_handler = AzureLogHandler(
                    connection_string=f"InstrumentationKey={self.instrumentation_key}"
                )
                if self.callback is not None:
                    log_handler.add_telemetry_processor(self.callback)
                logger.addHandler(log_handler)
            kwargs["logger"] = logger
        except Exception as ex:
            logging.error(f"Failed to create insights logger: {ex}")
            raise ex

        # execute main code
        try:
            # check for basics
            if self.send_basics:
                logger.info("Metric Setup Complete")

            # execute the main code
            return func(*args, **kwargs)
        except Exception as ex:
            # check for basics
            if self.send_basics:
                logger.exception("Function execution failed", exc_info=ex)

            # re-raise exception for outer loops
            raise ex



"""Json Serialization Decorator.

Copyright (c) 2023, Felix Geilert
"""

import json
from typing import Any, Dict, Tuple, Union

from azure.functions import HttpRequest
from functown.errors import RequestError

from .base import SerializationDecorator, DeserializationDecorator


class JsonResponse(SerializationDecorator):
    """Provides a JSON serialized response for an Azure Function.

    Args:
        func (Callable): The function to decorate.
        headers (Dict[str, str]): The headers to add to the response.
        status_code (int): The status code of the response.

    Example:
        >>> @JsonResponse
        ... def main(req: HttpRequest) -> Dict[str, str]:
        ...     return {"hello": "world"}

        >>> @JsonResponse(headers={"X-My-Header": "My-Value"})
        ... def main(req: HttpRequest) -> Dict[str, str]:
        ...     return {"hello": "world"}
    """

    def __init__(
        self,
        func=None,
        headers: Dict[str, str] = None,
        status_code: int = 200,
        **kwargs
    ):
        super().__init__(func, headers, status_code, **kwargs)

    def serialize(
        self, req: HttpRequest, res: Any, *args, **kwargs
    ) -> Tuple[Union[bytes, str], str]:
        return json.dumps(res), "application/json"


class JsonRequest(DeserializationDecorator):
    """Provides a JSON deserialized request for an Azure Function.

    Args:
        func (Callable): The function to decorate.
        enfore_mime (bool): Whether to enforce the mimetype check of the request body.
            Defaults to `True`.

    Example:
        >>> @JsonRequest
        ... def main(req: HttpRequest, body: dict):
        ...     item = body["item"]
    """

    def __init__(self, func=None, enfore_mime: bool = True, **kwargs):
        super().__init__(func, **kwargs)
        self._enforce = enfore_mime

    def deserialize(self, req: HttpRequest, *args, **kwargs) -> Any:
        # validate mimetype
        if (
            self._enforce is True
            and req.headers.get("Content-Type") != "application/json"
        ):
            raise RequestError("Request body must be JSON.", 400)

        # retrieve body and decode to string
        body = req.get_body()
        if isinstance(body, bytes):
            body = body.decode("utf-8")
        elif not isinstance(body, str):
            raise RequestError("Request body must be a string or bytes object.", 400)

        # handle loading
        return json.loads(body)



import threading


class ThreadSafeSingleton(type):
    _instances = {}
    _singleton_lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        # double-checked locking pattern (https://en.wikipedia.org/wiki/Double-checked_locking)
        if cls not in cls._instances:
            with cls._singleton_lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(ThreadSafeSingleton, cls).__call__(
                        *args, **kwargs
                    )
        return cls._instances[cls]



"""Decorator to automatically parse JWT Tokens from the request of the azure function.

Copyright (c) 2023, Felix Geilert
"""

import logging
from typing import List, Optional

from azure.functions import HttpRequest

from functown.utils import BaseDecorator
from .jwt import verify_user


class AuthHandler(BaseDecorator):
    """Decorator to automatically parse a JWT Token from the HttpRequest parameters.

    This will add an `token` parameter of type `Token` to the inner function signature.
    """

    def __init__(
        self,
        scopes: Optional[List[str]] = None,
        issuer_url: Optional[str] = None,
        audience: Optional[str] = None,
        verify: bool = True,
        auto_disable_verify: bool = True,
        **kwargs
    ):
        super().__init__(None, added_kw=["token"], **kwargs)

        self.scopes = scopes
        self.issuer_url = issuer_url
        self.audience = audience
        if verify and not issuer_url:
            if auto_disable_verify:
                logging.warning(
                    "No issuer_url provided, but verify is True. "
                    "Disabling token verification."
                )
                verify = False
            else:
                raise ValueError("No issuer_url provided, but verify is True.")
        self.verify = verify

    def run(self, func, *args, **kwargs):
        # retrieve request
        req: HttpRequest = self._get("req", 0, *args, **kwargs)

        # parse token
        token = verify_user(
            req,
            scopes=self.scopes,
            issuer_url=self.issuer_url,
            audience=self.audience,
            verify=self.verify,
        )

        kwargs["token"] = token

        return func(*args, **kwargs)
